
'''
该文件名为 `train_en_bert.py`，它是一个用于训练ENBert模型的程序文件。以下是它的主要功能和结构：

1. 导入所需的库和模块。
2. 设置环境变量，指定GPU设备。
3. 定义一些超参数：模型名称、最大单词长度、学习率、批大小、训练轮数等。
4. 定义一个缓存的装饰器函数，用于保存数据集并进行缓存。
5. 获取数据集并进行处理。
6. 创建模型并设置模型的参数。
7. 创建优化器。
8. 准备训练数据和验证数据。
9. 定义一些回调函数，用于在训练过程中进行一些操作。
10. 创建训练器并进行训练。

简而言之，该程序文件实现了使用ENBert模型进行训练的功能，包括数据处理、模型构建、优化器设置、训练数据准备和训练过程的设置。
'''

import sys
sys.path.append('../')

import os
if 'p' in os.environ:
	os.environ['CUDA_VISIBLE_DEVICES']=os.environ['p']

import torch
from torch import optim
from fastNLP import Trainer, CrossEntropyLoss
from fastNLP import BucketSampler, cache_results, WarmupCallback, GradientClipCallback, FitlogCallback
from mono.data.pipe import ENBertPipe
from mono.model.bert import ENBertReverseDict
import fitlog
from mono.model.metrics import MonoMetric
# from joint.data.utils import clip_max_length
# fitlog.debug()
fitlog.set_log_dir('en_logs')
fitlog.add_hyper_in_file(__file__)
fitlog.add_other('uncased', name='note')

# paths = '../data/en'
paths = '../data/biobricks'
#######hyper
model_name = 'bert'
# max_word_len = 5
max_word_len = 10 #** 观察所有word2bpes发现
lr = 2e-5
batch_size = 64
n_epochs = 10
#######hyper
# pre_name = 'bert-base-uncased1' #* 原
pre_name = 'biobricks-bert-base-uncased' #** me
# 在transformers中的名字叫做bert-base-cased


@cache_results('caches/en_{}_{}.pkl'.format(pre_name.split('/')[-1], max_word_len), _refresh=False) #* 此行注释后则不会报错cache文件
def get_data():
	data_bundle = ENBertPipe(pre_name, max_word_len).process_from_file(paths)
	return data_bundle

data_bundle = get_data() #* 获取数据, 会自动从mono/caches中读取
print('data_bundle:', data_bundle)

word2bpes = data_bundle.word2bpes
print(f"In total {len(word2bpes)} target words")
pad_id = data_bundle.pad_id

model = ENBertReverseDict(pre_name, word2bpes, pad_id=pad_id, #** 加载预训练模型
						number_word_in_train=data_bundle.number_word_in_train)

if torch.cuda.is_available():
	model.cuda()

optimizer = optim.AdamW(model.parameters(), lr=lr)

data = {}
# for name in ['seen', 'unseen', 'desc']: #* 原, 测试数据
for name in ['seen1-100']: #* me, 测试数据
	data[name] = data_bundle.get_dataset(name)

callbacks = [GradientClipCallback(clip_type='value', clip_value=5), WarmupCallback(warmup=0.01, schedule='linear')]
callbacks.append(FitlogCallback(data=data, verbose=1))

train_data = data_bundle.get_dataset('train') #* 训练数据
train_data.add_seq_len('input')

# from collections import Counter
# print(Counter(train_data.get_field('seq_len').content))
# exit(0)

sampler = BucketSampler()
# clip_max_length(train_data, data_bundle)

trainer = Trainer(train_data=train_data, 
				model=model,
				optimizer=optimizer, 
				loss=CrossEntropyLoss(),
				batch_size=64, # 64
				sampler=sampler, drop_last=False, update_every=1,
				num_workers=1, print_every=5,
				n_epochs=100, # 10
				dev_data=data_bundle.get_dataset('dev'), #* dev_data
				metrics=MonoMetric(),
				# metric_key='t10',
				metric_key='t1', #* 以此为指标决定保存最优模型
				validate_every=-1, save_path='save_models/', use_tqdm=True, 
				# device=None,
				device=5, #** 指定gpu设备,10服务器选0-6
				callbacks=callbacks, check_code_level=0)

trainer.train(load_best_model=False) #* 开始训练

fitlog.add_other(trainer.start_time, name='start_time')